/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mediator.DOMUtils;

import java.util.ArrayList;

import org.xquark.datastructure.EvaluationVisitor;
import org.xquark.datastructure.Tuple;
import org.xquark.mediator.plan.DynamicContext;
import org.xquark.xpath.datamodel.TypedDocumentImpl;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;

public class DOMEval {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	protected static TypedDocumentImpl docimpl = null;
	static {
		docimpl = new TypedDocumentImpl();
	}

	protected static EvaluationVisitor evalVisitor = null;
	static {
		evalVisitor = new EvaluationVisitor(null);
	}

	public static boolean evalCondition(DynamicContext context, XQueryExpression expr, Tuple tuple) throws XQueryException {
		try {
			evalVisitor.setSchemaManager(context.getSchemaManager());
			if (tuple != null)
				context.addCurrentTuple(tuple);
			evalVisitor.reset(context.getCurrentTuples());
			evalVisitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
			expr.accept(evalVisitor);
			if (tuple != null)
				context.deleteCurrentTuple(tuple);
			return evalVisitor.getVerdict();
		} catch (XQueryException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	public static ArrayList evalNodes(DynamicContext context, XQueryExpression expr, Tuple tuple) throws XQueryException {
		try {
			evalVisitor.setSchemaManager(context.getSchemaManager());
			if (tuple != null)
				context.addCurrentTuple(tuple);
			evalVisitor.reset(context.getCurrentTuples());
			evalVisitor.setReturnType(EvaluationVisitor.NODE_TYPE);
			expr.accept(evalVisitor);
			if (tuple != null)
				context.deleteCurrentTuple(tuple);
			return evalVisitor.getResNodes();
		} catch (XQueryException e) {
			throw new XQueryException(e.getMessage());
		}
	}

}
