/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mediator.decomposer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Stack;

import org.xquark.mediator.algebra.AlgArithFunction;
import org.xquark.mediator.algebra.AlgEval;
import org.xquark.mediator.algebra.AlgExtVar;
import org.xquark.mediator.algebra.AlgFLWR;
import org.xquark.mediator.algebra.AlgFunction;
import org.xquark.mediator.algebra.AlgLocated;
import org.xquark.mediator.algebra.AlgSequence;
import org.xquark.mediator.algebra.AlgSource;
import org.xquark.mediator.algebra.AlgSourceEval;
import org.xquark.mediator.algebra.AlgUnion;
import org.xquark.mediator.algebra.AlgVar;
import org.xquark.mediator.algebra.Algebra;
import org.xquark.mediator.algebra.AlgebraManager;
import org.xquark.xquery.parser.AggregateFunctionCall;
import org.xquark.xquery.parser.AttributeValuePair;
import org.xquark.xquery.parser.DefaultParserVisitor;
import org.xquark.xquery.parser.Element;
import org.xquark.xquery.parser.FLWRExpression;
import org.xquark.xquery.parser.LibraryFunctionCall;
import org.xquark.xquery.parser.ListOpArithExpression;
import org.xquark.xquery.parser.ListOpUNIONExpression;
import org.xquark.xquery.parser.LocatedExpression;
import org.xquark.xquery.parser.ParenthesizedExpression;
import org.xquark.xquery.parser.PrimitiveFunctionCall;
import org.xquark.xquery.parser.Step;
import org.xquark.xquery.parser.UnOpMinusExpression;
import org.xquark.xquery.parser.Value;
import org.xquark.xquery.parser.Variable;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.parser.XQueryExpressionSequence;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionCOLLECTION;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDOC;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.typing.TypeVisitor;

/**
 * This abstract represent a Node/Leaf use in the construction
 * of the dependancy trees.
 *
 */

/**
 * to be commented ... : objectif of this class
 * <code>AlgebraStructureVisitor</code> visit : 1. call
 * <code>DepManager</code> to create Dependence tree. 2.
 */

public class AlgebraStructureVisitor extends DefaultParserVisitor {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.15 $";

    private static final String RCSName = "Name";

    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************
    public static final Integer IN_LOCATED = new Integer(1);

    public static final Integer IN_DEFINITION = new Integer(2);

    public static final Integer IN_FLWR = new Integer(3);

    public static final Integer IN_WHERE = new Integer(4);

    public static final Integer IN_FUNCTION = new Integer(5);

    public static final Integer IN_MAIN_DEFINITION = new Integer(6);

    public static final Integer IN_SEQUENCE = new Integer(7);
    
    public static final Integer IN_EXT_VAR_DECL = new Integer(8);

    //public static final Integer IN_PREDICATE = new Integer(9);

    // elements passed at construction time
    private AlgebraManager depmanager = null;

    private HashMap parentvariable = null;

    // private temporary elements
    private Stack varStack = new Stack();

    private AlgFLWR currentDepFLWR = null;

    private Algebra currentDepNode = null;

    private TypeVisitor typevisitor = null;

    private Stack stateStack = new Stack();

    private boolean ongoingLet = false;

    private LocatedExpression ongoingLocatedExpression = null;

    //private ArrayList currentSources = null;

    // ************************************************************************
    // * INITIALIZATION
    // ************************************************************************
    public AlgebraStructureVisitor(AlgebraManager depmanager) throws XQueryException {
        this.depmanager = depmanager;
        this.typevisitor = depmanager.getTypeVisitor();
        this.parentvariable = new HashMap();
        //stateStack.push(IN_FLWR);
    }

    public void setStartState(Integer state) {
        stateStack.clear();
        stateStack.push(state);
    }
    // ************************************************************************
    // * GET/SET METHODS
    // ************************************************************************

    private void addLeaf(Algebra leaf) throws XQueryException {
        if (currentDepNode != null)
            currentDepNode.addLeaf(leaf);
        else if (currentDepFLWR != null)
            currentDepFLWR.addLeaf(leaf);
    }

    // ************************************************************************
    // * METHODS
    // ************************************************************************

    public void visit(AggregateFunctionCall arg) throws XQueryException {
        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgFunction depfunction = depmanager.addFunction(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, true);
            addLeaf(depfunction);
            ongoingLet = false;
            ArrayList arguments = arg.getArguments();
            if (arguments != null) {
                Algebra tmpDepNode = currentDepNode;
                currentDepNode = depfunction;
                stateStack.push(IN_FUNCTION);
                for (int i = 0; i < arguments.size(); i++)
                    arg.getArgument(i).accept(this);
                stateStack.pop();
                currentDepNode = tmpDepNode;
            }
        }
    }

    public void visit(LibraryFunctionCall arg) throws XQueryException {
        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgSource depsource = depmanager.addSource(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, ongoingLet);
            addLeaf(depsource);
            ongoingLet = false;
            //			ArrayList arguments = arg.getArguments();
            //			if (arguments != null) {
            //				Algebra tmpDepNode = currentDepNode;
            //				currentDepNode = depsource;
            //				stateStack.push(IN_FUNCTION);
            //				for (int i = 0; i < arguments.size(); i++)
            //					arg.getArgument(i).accept(this);
            //				stateStack.pop();
            //				currentDepNode = tmpDepNode;
            //			}
        }
    }

    public void visit(ListOpArithExpression arg) throws XQueryException {
        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgArithFunction deparithfunction = depmanager.addArithFunction(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, true);
            addLeaf(deparithfunction);
            ongoingLet = false;
            Algebra tmpDepNode = currentDepNode;
            currentDepNode = deparithfunction;
            stateStack.push(IN_FUNCTION);
            arg.getExpression1().accept(this);
            arg.getExpression2().accept(this);
            stateStack.pop();
            currentDepNode = tmpDepNode;
        }
    }

    public void visit(ListOpUNIONExpression arg) throws XQueryException {
        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgUnion depunion = depmanager.addUnion(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, true);
            addLeaf(depunion);
            ongoingLet = false;
            Algebra tmpDepNode = currentDepNode;
            currentDepNode = depunion;
            stateStack.push(IN_FUNCTION);
            arg.getExpression1().accept(this);
            arg.getExpression2().accept(this);
            stateStack.pop();
            currentDepNode = tmpDepNode;
        }
    }

    public void visit(FLWRExpression arg) throws XQueryException {

        ArrayList variables = arg.getVariables();
        Integer currentstate = (Integer) stateStack.peek();
        //currentSources = null;

        Algebra tmpDepNode = currentDepNode;
        if (currentstate == IN_DEFINITION) {
            AlgFLWR depflwr = depmanager.addFlwr(arg, (Variable) varStack.peek(), ongoingLet);
            addLeaf(depflwr);
            currentDepNode = depflwr;
            ongoingLet = false;
        } else if (currentstate == IN_FUNCTION) {
            AlgFLWR depflwr = depmanager.addFlwr(arg, null, ongoingLet);
            addLeaf(depflwr);
            currentDepNode = depflwr;
            ongoingLet = false;
        }

        // TEST if sources is only one
        /*
         * // if (sources != null && sources.size() == 1) { // for (int i=0;i
         * <variables.size();i++) { // Variable vari =
         * (Variable)variables.get(i); // Variable parentvari =
         * (Variable)parentvariable.get(vari); // if (parentvari != null) { //
         * currentDepMem.setDependency((DepMem)
         * depmanager.getMapNode(parentvari)); // variables.remove(i); // i--; // } //
         * else { // vari.setParentFLWRExpression(arg); //
         * parentvariable.put(vari,varStack.firstElement()); // } // } //
         * //Utils.fillSources(arg,new
         * UnitEnvironment(),this.depmanager.getMetaData()); // DepNodeSource
         * newsource = depmanager.addNodeSource(arg,arg.getVariables(),
         * ongoingLet); // addLeaf(newsource); //
         * this.currentDepMem.setMonoSource(true); // return; // }
         */
        /*
         * ArrayList removedVariables = new ArrayList(); for (int i=0;i
         * <variables.size();i++) { Variable vari = (Variable)variables.get(i);
         * Variable parentvari = (Variable)parentvariable.get(vari); if
         * (parentvari != null) { currentDepFLWR.setDependency((DepFLWR)
         * depmanager.getMapNode(parentvari)); removedVariables.add(vari);
         * variables.remove(i); i--; } } // remove conditions from where clause
         * implicating only removed variables if (arg.getWhereClause() != null &&
         * !removedVariables.isEmpty()) { IsolateWhereVisitor visitor = new
         * IsolateWhereVisitor(removedVariables) ; try {
         * arg.getWhereClause().accept(visitor) ; } catch (XQueryException e) {
         * throw new MediatorException("Can't isolate where expression : " +
         * arg.getWhereClause().pprint(), e) ; }
         * arg.setWhereClause(visitor.getRemainingExpression()) ; }
         */
        // TEST if sources is only one
        /*
         * Utils.fillSources(arg,this.depmanager.getMetaData());
         * 
         * ArrayList sources = arg.getSources(); if (sources != null &&
         * sources.size() == 1) { for (int i=0;i <variables.size();i++) {
         * Variable vari = (Variable)variables.get(i);
         * vari.setParentFLWRExpression(arg);
         * parentvariable.put(vari,varStack.firstElement()); }
         * //Utils.fillSources(arg,new
         * UnitEnvironment(),this.depmanager.getMetaData()); DepNodeSource
         * newsource = depmanager.addNodeSource(arg,arg.getVariables(),
         * ongoingLet); addLeaf(newsource); if (currentstate ==
         * IN_MAIN_DEFINITION) currentDepFLWR.setMonoSource(true); else if
         * (currentstate == IN_DEFINITION) currentDepNode = tmpDepNode; return; }
         */
        for (int i = 0; i < variables.size(); i++) {
            Variable vari = (Variable) variables.get(i);
            vari.setParentFLWRExpression(arg);
            stateStack.push(IN_FLWR);
            vari.accept(this);
            stateStack.pop();
        }

        //		ArrayList varsOnTheSameSource = new ArrayList();
        //		String lastSource = null;
        //		ArrayList sourcesOfvari = null;
        //		for (int i = 0; i < variables.size(); i++) {
        //			Variable vari = (Variable)variables.get(i);
        //			vari.setParentFLWRExpression(arg);
        //			sourcesOfvari = vari.getSources();
        //
        //			boolean isMonoSource = (null != sourcesOfvari && 1 ==
        // sourcesOfvari.size());
        //			if (isMonoSource && sourcesOfvari.get(0).equals(lastSource)) {
        //				/* vari is defined on the same source as the precedant variable*/
        //				varsOnTheSameSource.add(vari);
        //			}
        //			else {
        //				if (!varsOnTheSameSource.isEmpty()) {
        //					/* Generate DepNodeSource for accumulated variables*/
        //					DepNodeSource newsource =
        // depmanager.addNodeSource(arg,varsOnTheSameSource, ongoingLet);
        //					addLeaf(newsource);
        //
        //					varsOnTheSameSource = new ArrayList();
        //
        //				}
        //				/**/
        //				if (isMonoSource) {
        //					lastSource = (String)sourcesOfvari.get(0);
        //					varsOnTheSameSource.add(vari);
        //				}
        //				else {
        //					stateStack.push(IN_FLWR) ;
        //					vari.accept(this);
        //					stateStack.pop() ;
        //				}
        //			}
        //		}
        //
        //		/* treat last group of variables if there is */
        //		if (!varsOnTheSameSource.isEmpty()) {
        //			/* Generate DepNodeSource for accumulated variables*/
        //			DepNodeSource newsource =
        // depmanager.addNodeSource(arg,varsOnTheSameSource, ongoingLet);
        //			addLeaf(newsource);
        //		}

        XQueryExpression whereclause = arg.getWhereClause();
        if (whereclause != null) {
            stateStack.push(IN_WHERE);
            whereclause.accept(this);
            stateStack.pop();
        }

        //ArrayList orderbys = arg.getOrderBy() ;
        if (currentstate == IN_DEFINITION)
            currentDepNode = tmpDepNode;

    }

    public void visit(FunctionCOLLECTION arg) throws XQueryException {

        Integer state = (Integer) stateStack.peek();
        AlgSource newsource = null;
        if (state == IN_DEFINITION)
            newsource = depmanager.addSource(arg, (Variable) varStack.peek(), ongoingLet);
        else if (state == IN_LOCATED)
            newsource = depmanager.addSource(ongoingLocatedExpression, (Variable) null, ongoingLet);
        else
            newsource = depmanager.addSource(arg, (Variable) null, ongoingLet);
        //if (!ongoingLet)
        addLeaf(newsource);
        ongoingLet = false;
    }

    public void visit(FunctionDOC arg) throws XQueryException {

        Integer state = (Integer) stateStack.peek();
        AlgSourceEval newsource = null;
        if (state == IN_DEFINITION)
            newsource = depmanager.addSourceEval(arg, (Variable) varStack.peek(), ongoingLet);
        else if (state == IN_FUNCTION)
            newsource = depmanager.addSourceEval(arg, (Variable) varStack.peek(), ongoingLet);
        else if (state == IN_LOCATED)
            newsource = depmanager.addSourceEval(ongoingLocatedExpression, (Variable) null, ongoingLet);
        else
            newsource = depmanager.addSourceEval(arg, (Variable) null, ongoingLet);
        //if (!ongoingLet)
        addLeaf(newsource);
        ongoingLet = false;
    }

    //	public void visit(ExternalVariable arg) throws XQueryException {
    //		Integer state = (Integer) stateStack.peek();
    //		AlgValue newExtVar =
    // depmanager.addValue(null,(Variable)arg.getObjectValue(),true);
    //		addLeaf(newExtVar);
    //		ongoingLet = false;
    //	}

    public void visit(PrimitiveFunctionCall arg) throws XQueryException {

        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgFunction depfunction = depmanager.addFunction(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, ongoingLet);
            addLeaf(depfunction);
            ongoingLet = false;
            ArrayList arguments = arg.getArguments();
            if (arguments != null) {
                Algebra tmpDepNode = currentDepNode;
                currentDepNode = depfunction;
                stateStack.push(IN_FUNCTION);
                for (int i = 0; i < arguments.size(); i++)
                    arg.getArgument(i).accept(this);
                stateStack.pop();
                currentDepNode = tmpDepNode;
            }
        }

    }

    public void visit(LocatedExpression arg) throws XQueryException {
        Integer state = (Integer) stateStack.peek();
        ArrayList steps = arg.getSteps();
        if (state != IN_DEFINITION)
            stateStack.push(IN_LOCATED);
        ongoingLocatedExpression = arg;
        XQueryExpression exp = ((Step) steps.get(0)).getExpression();
        exp.accept(this);
        if (state != IN_DEFINITION)
            stateStack.pop();
        ongoingLocatedExpression = null;
    }

    public void visit(Value arg) throws XQueryException {
        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgEval depeval = depmanager.addEval(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, true);
            addLeaf(depeval);
            ongoingLet = false;
        }
    }

    private GetUsedVariablesVisitor guvvisitor = null;

    public void visit(Variable arg) throws XQueryException {
        Integer currentstate = (Integer) stateStack.peek();

        if (currentstate == IN_EXT_VAR_DECL && arg.isExternal()) {
            Integer state = (Integer) stateStack.peek();
            AlgExtVar newExtVar = depmanager.addExtVar(arg);
            addLeaf(newExtVar);
            ongoingLet = false;
            return;
        }
        if (currentstate == IN_DEFINITION) {
            if (arg.getExpression() == null && !arg.isExternal()) {
                AlgSource depsource = depmanager.addSource(arg, (Variable) varStack.peek(), ongoingLet);
                addLeaf(depsource);
                ongoingLet = false;
            } else if (ongoingLocatedExpression == null) {
                AlgVar depvar = depmanager.addVar(arg, (Variable) varStack.peek(), ongoingLet);
                addLeaf(depvar);
                ongoingLet = false;
            } else {
                AlgLocated deplocated = depmanager.addLocated(ongoingLocatedExpression, arg, (Variable) varStack.peek(), ongoingLet);
                addLeaf(deplocated);
                ongoingLet = false;
            }
        } else if (currentstate == IN_FLWR) {
            if (arg.getBindingType() == Constants.LET_BINDINGTYPE) {
                ongoingLet = true;
            }
            varStack.push(arg);
            if (arg.getParentFLWRExpression() == null) {
                // remove variables if necessary
                FLWRExpression tmpflwr = (FLWRExpression) arg.getExpression();
                ArrayList variables = tmpflwr.getVariables();
                Variable lastremovedvar = null;
                Variable lastparentvar = null;
                for (int i = 0; i < variables.size(); i++) {
                    Variable vari = (Variable) variables.get(i);
                    Variable parentvari = (Variable) parentvariable.get(vari);
                    if (parentvari == null)
                        break;
                    lastremovedvar = vari;
                    if (parentvari != lastparentvar) {
                        lastparentvar = parentvari;
                        if (tmpflwr.getWhereClause() != null) {
                            XQueryExpression tmpwhere = ((FLWRExpression) depmanager.getMapNode(lastparentvar).getExpression()).getWhereClause();
                            if (tmpflwr.getWhereClause() != null)
                                tmpflwr.setWhereClause(tmpflwr.getWhereClause().erasePart(tmpwhere));
                        }
                    }
                    variables.remove(i);
                    i--;
                }

                currentDepFLWR = depmanager.addQDB(arg, ongoingLet);
                if (lastremovedvar != null)
                    currentDepFLWR.setDependency((AlgFLWR) depmanager.getMapNode((Variable) parentvariable.get(lastremovedvar)));
                if (currentDepFLWR.getVarsDependingOn() != null)
                    currentDepFLWR.updateVarDependencyList((ArrayList) currentDepFLWR.getVarsDependingOn().clone());

                ongoingLet = false;
                stateStack.push(IN_MAIN_DEFINITION);
            } else {
                parentvariable.put(arg, varStack.firstElement());
                stateStack.push(IN_DEFINITION);
            }
            arg.getExpression().accept(this);
            stateStack.pop();
            varStack.pop();
        }
    }

    public void visit(UnOpMinusExpression arg) throws XQueryException {

        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgFunction depfunction = depmanager.addFunction(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, ongoingLet);
            addLeaf(depfunction);
            ongoingLet = false;
            Algebra tmpDepNode = currentDepNode;
            currentDepNode = depfunction;
            stateStack.push(IN_FUNCTION);
            arg.getExpression().accept(this);
            stateStack.pop();
            currentDepNode = tmpDepNode;
        }
    }

    public void visit(XQueryExpressionSequence arg) throws XQueryException {

        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION /* || state == IN_FUNCTION */
        ) {
            AlgSequence depsequence = depmanager.addSequence(arg, (state == IN_DEFINITION) ? (Variable) varStack.peek() : null, ongoingLet);
            addLeaf(depsequence);
            ongoingLet = false;
            ArrayList expressions = arg.getSubExpressions();
            if (expressions != null) {
                stateStack.push(IN_FUNCTION);
                for (int i = 0; i < expressions.size(); i++) {
                    Algebra tmpDepNode = currentDepNode;
                    currentDepNode = depsequence;
                    ((XQueryExpression) expressions.get(i)).accept(this);
                    currentDepNode = tmpDepNode;
                }
                stateStack.pop();
            }
        }
    }

    public void visit(ParenthesizedExpression arg) throws XQueryException {
        if (arg.getExpression() != null)
            arg.getExpression().accept(this);
    }

    // add LARS 16/09/2003
    public void visit(Element arg) throws XQueryException {

        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION) {
            AlgEval depeval = null;
            if (ongoingLocatedExpression == null) {
                depeval = depmanager.addEval(arg, (Variable) varStack.peek(), ongoingLet);
            } else {
                depeval = depmanager.addEval(ongoingLocatedExpression, (Variable) varStack.peek(), ongoingLet);
            }
            addLeaf(depeval);
        }
    }

    public void visit(AttributeValuePair arg) throws XQueryException {

        Integer state = (Integer) stateStack.peek();
        if (state == IN_DEFINITION) {
            AlgEval depeval = depmanager.addEval(arg, (Variable) varStack.peek(), ongoingLet);
            addLeaf(depeval);
        }
    }

    public void visit(XQueryExpression arg) throws XQueryException {
        //throw new XQueryException("CALL AlgebraStructureVisitor for
        // XQueryExpression : " + arg.getClass().getName());
    }

}